package biz.hammurapi.swing;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Enumeration;
import java.util.List;

import javax.swing.tree.TreeNode;

import biz.hammurapi.util.CollectionVisitable;
import biz.hammurapi.util.Visitable;
import biz.hammurapi.util.Visitor;

/**
 * Base class for tree nodes which load their children on demand.
 * Subclasses shall implement loadChildren() method.
 * @author Pavel
 *
 */
public abstract class LazyTreeNode implements TreeNode, Visitable {
	
	private List children;
	
	protected String name;

	protected TreeNode parent;
	
	protected String description;
	
	protected static final List EMPTY_LIST = Collections.unmodifiableList(new ArrayList());
	
	public String toString() {
		return name;
	}
	
	public LazyTreeNode(TreeNode parent, String name) {
		this(parent);
		this.name = name;
	}
	
	protected LazyTreeNode(TreeNode parent, String name, List children) {
		this(parent, name);
		this.children = children;
	}	
	
	public LazyTreeNode(TreeNode parent) {
		this.parent = parent;
	}
	
	/**
	 * Visits only children if they are loaded.
	 */
	public boolean accept(Visitor visitor) {
		if (visitor.visit(this)) {
			if (children!=null) {
				new CollectionVisitable(children, false).accept(visitor);
			}
			return true;
		}
		return false;
	}
	
	protected List getChildren() {
		if (children==null) {
			children = loadChildren();
			if (children == null) {
				children = EMPTY_LIST;
			}
		}
		return children;
	}

	/**
	 * Loads node children.
	 * @return
	 */
	protected abstract List loadChildren();

	public Enumeration children() {
		return Collections.enumeration(getChildren());
	}

	public boolean getAllowsChildren() {
		return true;
	}

	public TreeNode getChildAt(int childIndex) {
		return (TreeNode) getChildren().get(childIndex);
	}

	public int getChildCount() {
		return getChildren().size();
	}

	public int getIndex(TreeNode node) {
		return getChildren().indexOf(node);
	}

	public TreeNode getParent() {
		return parent;
	}

	public boolean isLeaf() {
		return getChildren().isEmpty();
	}
		
	public boolean isSelectable() {
		return true;
	}
	
	/**
	 * @return Text for rendering tooltip.
	 */
	public String getDescription() {
		return description;
	}
}
