package biz.hammurapi.util;

import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.lang.reflect.UndeclaredThrowableException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

public class VisitorConverter {
	
	/**
	 * Name of leave method.
	 */
	public static final String LEAVE = "leave";
	
	/**
	 * Name of visit method.
	 */
	public static final String VISIT = "visit";
	
	/**
	 * Converts object to visitor.
	 * If visitor is instance of <code>Visitor</code>, then this method returns
	 * object unchanged. Otherwise it introspects object for public 
	 * <boolean>visit()</boolean> and <boolean>leave()</boolean> methods
	 * with a single argument returning boolean or void. If such methods are found
	 * they are made accessible in order to enable invocation of methods declared in
	 * anonymous and local classes. This method returns an anonymous visitor or polite
	 * visitor (if there are leave() methods). This anonymous (polite) visitor
	 * dispatches invocations to visit() and leave() methods with compatible types. 
	 * @param target Object to be converted to visitor
	 * @return Object argument if it already implements visitor, or dispatcher 
	 * to object's visit() and leave() methods.
	 */
	public Visitor convert(final Object object) {
		if (object == null ) {
			return null;
		}
		
		if (object instanceof Visitor) {
			return (Visitor) object;
		}		
		
		final List visitMethods = new ArrayList();
		final List leaveMethods = new ArrayList();

		Method[] methods = object.getClass().getMethods();
		for (int i=0; i<methods.length; ++i) {
			Method method = methods[i];
			if (Modifier.isPublic(method.getModifiers())
					&& VISIT.equals(method.getName()) 
					&& method.getParameterTypes().length==1
					&& (void.class.equals(method.getReturnType()) || boolean.class.equals(method.getReturnType()))) {
				
				if (!method.isAccessible()) {
					method.setAccessible(true);
				}
				visitMethods.add(method);
			}
			
			if (Modifier.isPublic(method.getModifiers()) 
					&& LEAVE.equals(method.getName()) 
					&& method.getParameterTypes().length==1
					&& void.class.equals(method.getReturnType())) {
				
				if (!method.isAccessible()) {
					method.setAccessible(true);
				}
				leaveMethods.add(method);
			}
		}
		
		if (visitMethods.isEmpty() && leaveMethods.isEmpty()) {
			return null;
		}
		
		final Visitor dispatcher = new Visitor() {

			public boolean visit(Object target) {
				Iterator it = visitMethods.iterator();
				while (it.hasNext()) {
					Method visitMethod = (Method) it.next();
					if (visitMethod.getParameterTypes()[0].isInstance(target)) {
						try {
							Object ret = visitMethod.invoke(object, new Object[] {target});
							if (Boolean.FALSE.equals(ret)) {
								return false;
							}
						} catch (Exception e) {
							throw new UndeclaredThrowableException(e);
						}
					}
				}
				
				return true;
			}
			
		};
		
		if (leaveMethods.isEmpty()) {
			return dispatcher;
		}
		
		return new PoliteVisitor() {

			public void leave(Object target) {
				Iterator it = visitMethods.iterator();
				while (it.hasNext()) {
					Method visitMethod = (Method) it.next();
					if (visitMethod.getParameterTypes()[0].isInstance(target)) {
						try {
							visitMethod.invoke(object, new Object[] {target});
						} catch (Exception e) {
							throw new UndeclaredThrowableException(e);
						}
					}
				}
			}

			public boolean visit(Object target) {
				return dispatcher.visit(target);
			}
			
		};
	}

}
